@@ -27,7 +27,6 @@
 
 #include "config.h"
 
-#include <glib/gi18n-lib.h>
 #include <math.h>
 #include <string.h>
 
@@ -70,6 +69,8 @@ static GParamSpec *manager_properties[LAST_PROP] = { NULL };
 
 enum
 {
+  MONITORS_CHANGED,
+  POWER_SAVE_MODE_CHANGED,
   CONFIRM_DISPLAY_CHANGE,
 
   LAST_SIGNAL
@@ -77,11 +78,7 @@ enum
 
 static guint manager_signals[LAST_SIGNAL] = { 0 };
 
-static void gf_monitor_manager_display_config_init (GfDBusDisplayConfigIface *iface);
-
-G_DEFINE_ABSTRACT_TYPE_WITH_CODE (GfMonitorManager, gf_monitor_manager, GF_DBUS_TYPE_DISPLAY_CONFIG_SKELETON,
-                                  G_ADD_PRIVATE (GfMonitorManager)
-                                  G_IMPLEMENT_INTERFACE (GF_DBUS_TYPE_DISPLAY_CONFIG, gf_monitor_manager_display_config_init))
+G_DEFINE_TYPE_WITH_PRIVATE (GfMonitorManager, gf_monitor_manager, G_TYPE_OBJECT)
 
 /* Array index matches GfMonitorTransform */
 static gfloat transform_matrices[][6] =
@@ -137,12 +134,10 @@ power_save_mode_changed (GfMonitorManager *manager,
                          gpointer          user_data)
 {
   GfMonitorManagerClass *manager_class;
-  GfDBusDisplayConfig *display_config;
   gint mode;
 
   manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
-  display_config = GF_DBUS_DISPLAY_CONFIG (manager);
-  mode = gf_dbus_display_config_get_power_save_mode (display_config);
+  mode = gf_dbus_display_config_get_power_save_mode (manager->display_config);
 
   if (mode == GF_POWER_SAVE_UNSUPPORTED)
     return;
@@ -150,7 +145,8 @@ power_save_mode_changed (GfMonitorManager *manager,
   /* If DPMS is unsupported, force the property back. */
   if (manager->power_save_mode == GF_POWER_SAVE_UNSUPPORTED)
     {
-      gf_dbus_display_config_set_power_save_mode (display_config, GF_POWER_SAVE_UNSUPPORTED);
+      gf_dbus_display_config_set_power_save_mode (manager->display_config,
+                                                  GF_POWER_SAVE_UNSUPPORTED);
       return;
     }
 
@@ -158,6 +154,8 @@ power_save_mode_changed (GfMonitorManager *manager,
     manager_class->set_power_save_mode (manager, mode);
 
   manager->power_save_mode = mode;
+
+  g_signal_emit (manager, manager_signals[POWER_SAVE_MODE_CHANGED], 0);
 }
 
 static void
@@ -180,11 +178,11 @@ gf_monitor_manager_notify_monitors_changed (GfMonitorManager *manager)
 
   priv = gf_monitor_manager_get_instance_private (manager);
 
-  manager->current_switch_config = GF_MONITOR_SWITCH_CONFIG_UNKNOWN;
-
   gf_backend_monitors_changed (priv->backend);
 
-  g_signal_emit_by_name (manager, "monitors-changed");
+  g_signal_emit (manager, manager_signals[MONITORS_CHANGED], 0);
+
+  gf_dbus_display_config_emit_monitors_changed (manager->display_config);
 }
 
 static GfMonitor *
@@ -206,6 +204,12 @@ find_monitor (GfMonitorManager *monitor_manager,
   return NULL;
 }
 
+static GfMonitor *
+get_active_monitor (GfMonitorManager *manager)
+{
+  return find_monitor (manager, gf_monitor_is_active);
+}
+
 static gboolean
 gf_monitor_manager_is_config_applicable (GfMonitorManager  *manager,
                                          GfMonitorsConfig  *config,
@@ -322,13 +326,17 @@ calculate_monitor_scale (GfMonitorManager *manager,
 static gfloat
 derive_calculated_global_scale (GfMonitorManager *manager)
 {
-  GfMonitor *primary_monitor;
+  GfMonitor *monitor;
 
-  primary_monitor = gf_monitor_manager_get_primary_monitor (manager);
-  if (!primary_monitor)
+  monitor = gf_monitor_manager_get_primary_monitor (manager);
+
+  if (!monitor || !gf_monitor_is_active (monitor))
+    monitor = get_active_monitor (manager);
+
+  if (!monitor)
     return 1.0;
 
-  return calculate_monitor_scale (manager, primary_monitor);
+  return calculate_monitor_scale (manager, monitor);
 }
 
 static GfLogicalMonitor *
@@ -935,93 +943,6 @@ is_valid_layout_mode (GfLogicalMonitorLayoutMode layout_mode)
   return FALSE;
 }
 
-static const gdouble known_diagonals[] =
-  {
-    12.1,
-    13.3,
-    15.6
-  };
-
-static gchar *
-diagonal_to_str (gdouble d)
-{
-  guint i;
-
-  for (i = 0; i < G_N_ELEMENTS (known_diagonals); i++)
-    {
-      gdouble delta;
-
-      delta = fabs(known_diagonals[i] - d);
-
-      if (delta < 0.1)
-        return g_strdup_printf ("%0.1lf\"", known_diagonals[i]);
-    }
-
-  return g_strdup_printf ("%d\"", (int) (d + 0.5));
-}
-
-static gchar *
-make_display_name (GfMonitorManager *manager,
-                   GfOutput         *output)
-{
-  gchar *inches;
-  gchar *vendor_name;
-
-  if (gf_output_is_laptop (output))
-    return g_strdup (_("Built-in display"));
-
-  inches = NULL;
-  vendor_name = NULL;
-
-  if (output->width_mm > 0 && output->height_mm > 0)
-    {
-      gint width_mm;
-      gint height_mm;
-      gdouble d;
-
-      width_mm = output->width_mm;
-      height_mm = output->height_mm;
-      d = sqrt (width_mm * width_mm + height_mm * height_mm);
-
-      inches = diagonal_to_str (d / 25.4);
-    }
-
-  if (g_strcmp0 (output->vendor, "unknown") != 0)
-    {
-      if (!manager->pnp_ids)
-        manager->pnp_ids = gnome_pnp_ids_new ();
-
-      vendor_name = gnome_pnp_ids_get_pnp_id (manager->pnp_ids, output->vendor);
-
-      if (!vendor_name)
-        vendor_name = g_strdup (output->vendor);
-    }
-  else
-    {
-      if (inches != NULL)
-        vendor_name = g_strdup (_("Unknown"));
-      else
-        vendor_name = g_strdup (_("Unknown Display"));
-    }
-
-  if (inches != NULL)
-    {
-      gchar *display_name;
-
-      /* TRANSLATORS: this is a monitor vendor name, followed by a
-       * size in inches, like 'Dell 15"'
-       */
-      display_name = g_strdup_printf (_("%s %s"), vendor_name, inches);
-
-      g_free (vendor_name);
-      g_free (inches);
-
-      return display_name;
-    }
-
-  return vendor_name;
-}
-
 static const gchar *
 get_connector_type_name (GfConnectorType connector_type)
 {
@@ -1067,36 +988,62 @@ rebuild_monitors (GfMonitorManager *manager)
       manager->monitors = NULL;
     }
 
-  for (l = manager->outputs; l; l = l->next)
+  for (l = manager->gpus; l; l = l->next)
     {
-      GfOutput *output = l->data;
+      GfGpu *gpu = l->data;
+      GList *k;
 
-      if (output->tile_info.group_id)
+      for (k = gf_gpu_get_outputs (gpu); k; k = k->next)
         {
-          if (is_main_tiled_monitor_output (output))
+          GfOutput *output = k->data;
+
+          if (output->tile_info.group_id)
             {
-              GfMonitorTiled *monitor_tiled;
+              if (is_main_tiled_monitor_output (output))
+                {
+                  GfMonitorTiled *monitor_tiled;
 
-              monitor_tiled = gf_monitor_tiled_new (manager, output);
-              manager->monitors = g_list_append (manager->monitors, monitor_tiled);
+                  monitor_tiled = gf_monitor_tiled_new (gpu, output);
+                  manager->monitors = g_list_append (manager->monitors, monitor_tiled);
+                }
             }
-        }
-      else
-        {
-          GfMonitorNormal *monitor_normal;
+          else
+            {
+              GfMonitorNormal *monitor_normal;
 
-          monitor_normal = gf_monitor_normal_new (manager, output);
-          manager->monitors = g_list_append (manager->monitors, monitor_normal);
+              monitor_normal = gf_monitor_normal_new (gpu, output);
+              manager->monitors = g_list_append (manager->monitors, monitor_normal);
+            }
         }
     }
 }
 
+static GList *
+combine_gpu_lists (GfMonitorManager *manager,
+                   GList            * (*list_getter) (GfGpu *gpu))
+{
+  GList *list = NULL;
+  GList *l;
+
+  for (l = manager->gpus; l; l = l->next)
+    {
+      GfGpu *gpu = l->data;
+
+      list = g_list_concat (list, g_list_copy (list_getter (gpu)));
+    }
+
+  return list;
+}
+
 static gboolean
 gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
-                                         GDBusMethodInvocation *invocation)
+                                         GDBusMethodInvocation *invocation,
+                                         GfMonitorManager      *manager)
 {
-  GfMonitorManager *manager;
   GfMonitorManagerClass *manager_class;
+  GList *combined_modes;
+  GList *combined_outputs;
+  GList *combined_crtcs;
   GVariantBuilder crtc_builder;
   GVariantBuilder output_builder;
   GVariantBuilder mode_builder;
@@ -1105,14 +1052,17 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
   gint max_screen_width;
   gint max_screen_height;
 
-  manager = GF_MONITOR_MANAGER (skeleton);
-  manager_class = GF_MONITOR_MANAGER_GET_CLASS (skeleton);
+  manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
+
+  combined_modes = combine_gpu_lists (manager, gf_gpu_get_modes);
+  combined_outputs = combine_gpu_lists (manager, gf_gpu_get_outputs);
+  combined_crtcs = combine_gpu_lists (manager, gf_gpu_get_crtcs);
 
   g_variant_builder_init (&crtc_builder, G_VARIANT_TYPE ("a(uxiiiiiuaua{sv})"));
   g_variant_builder_init (&output_builder, G_VARIANT_TYPE ("a(uxiausauaua{sv})"));
   g_variant_builder_init (&mode_builder, G_VARIANT_TYPE ("a(uxuudu)"));
 
-  for (l = manager->crtcs, i = 0; l; l = l->next, i++)
+  for (l = combined_crtcs, i = 0; l; l = l->next, i++)
     {
       GfCrtc *crtc = l->data;
       GVariantBuilder transforms;
@@ -1124,7 +1074,7 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
           g_variant_builder_add (&transforms, "u", j);
 
       if (crtc->current_mode)
-        current_mode_index = g_list_index (manager->modes, crtc->current_mode);
+        current_mode_index = g_list_index (combined_modes, crtc->current_mode);
       else
         current_mode_index = -1;
 
@@ -1141,12 +1091,12 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
                              NULL /* properties */);
     }
 
-  for (l = manager->outputs, i = 0; l; l = l->next, i++)
+  for (l = combined_outputs, i = 0; l; l = l->next, i++)
     {
       GfOutput *output = l->data;
       GVariantBuilder crtcs, modes, clones, properties;
       GBytes *edid;
-      gchar *edid_file;
+      GfCrtc *crtc;
       gint crtc_index;
 
       g_variant_builder_init (&crtcs, G_VARIANT_TYPE ("au"));
@@ -1156,7 +1106,7 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
           guint possible_crtc_index;
 
           possible_crtc = output->possible_crtcs[j];
-          possible_crtc_index = g_list_index (manager->crtcs, possible_crtc);
+          possible_crtc_index = g_list_index (combined_crtcs, possible_crtc);
 
           g_variant_builder_add (&crtcs, "u", possible_crtc_index);
         }
@@ -1166,7 +1116,7 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
         {
           guint mode_index;
 
-          mode_index = g_list_index (manager->modes, output->modes[j]);
+          mode_index = g_list_index (combined_modes, output->modes[j]);
           g_variant_builder_add (&modes, "u", mode_index);
 
         }
@@ -1176,7 +1126,7 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
         {
           guint possible_clone_index;
 
-          possible_clone_index = g_list_index (manager->outputs,
+          possible_clone_index = g_list_index (combined_outputs,
                                                output->possible_clones[j]);
 
           g_variant_builder_add (&clones, "u", possible_clone_index);
@@ -1194,7 +1144,7 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
       g_variant_builder_add (&properties, "{sv}", "height-mm",
                              g_variant_new_int32 (output->height_mm));
       g_variant_builder_add (&properties, "{sv}", "display-name",
-                             g_variant_new_take_string (make_display_name (manager, output)));
+                             g_variant_new_string (output->name));
       g_variant_builder_add (&properties, "{sv}", "backlight",
                              g_variant_new_int32 (output->backlight));
       g_variant_builder_add (&properties, "{sv}", "min-backlight-step",
@@ -1211,23 +1161,14 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
       g_variant_builder_add (&properties, "{sv}", "supports-underscanning",
                              g_variant_new_boolean (output->supports_underscanning));
 
-      edid_file = manager_class->get_edid_file (manager, output);
-      if (edid_file)
-        {
-          g_variant_builder_add (&properties, "{sv}", "edid-file",
-                                 g_variant_new_take_string (edid_file));
-        }
-      else
-        {
-          edid = manager_class->read_edid (manager, output);
+      edid = manager_class->read_edid (manager, output);
 
-          if (edid)
-            {
-              g_variant_builder_add (&properties, "{sv}", "edid",
-                                     g_variant_new_from_bytes (G_VARIANT_TYPE ("ay"),
-                                                               edid, TRUE));
-              g_bytes_unref (edid);
-            }
+      if (edid)
+        {
+          g_variant_builder_add (&properties, "{sv}", "edid",
+                                 g_variant_new_from_bytes (G_VARIANT_TYPE ("ay"),
+                                                           edid, TRUE));
+          g_bytes_unref (edid);
         }
 
       if (output->tile_info.group_id)
@@ -1244,7 +1185,8 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
                                                 output->tile_info.tile_h));
         }
 
-      crtc_index = output->crtc ? g_list_index (manager->crtcs, output->crtc) : -1;
+      crtc = gf_output_get_assigned_crtc (output);
+      crtc_index = crtc ? g_list_index (combined_crtcs, crtc) : -1;
 
       g_variant_builder_add (&output_builder, "(uxiausauaua{sv})",
                              i, /* ID */
@@ -1257,7 +1199,7 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
                              &properties);
     }
 
-  for (l = manager->modes, i = 0; l; l = l->next, i++)
+  for (l = combined_modes, i = 0; l; l = l->next, i++)
     {
       GfCrtcMode *mode = l->data;
 
@@ -1285,6 +1227,10 @@ gf_monitor_manager_handle_get_resources (GfDBusDisplayConfig   *skeleton,
                                                  g_variant_builder_end (&mode_builder),
                                                  max_screen_width, max_screen_height);
 
+  g_list_free (combined_modes);
+  g_list_free (combined_outputs);
+  g_list_free (combined_crtcs);
+
   return TRUE;
 }
 
@@ -1293,14 +1239,14 @@ gf_monitor_manager_handle_change_backlight (GfDBusDisplayConfig   *skeleton,
                                             GDBusMethodInvocation *invocation,
                                             guint                  serial,
                                             guint                  output_index,
-                                            gint                   value)
+                                            gint                   value,
+                                            GfMonitorManager      *manager)
 {
-  GfMonitorManager *manager;
   GfMonitorManagerClass *manager_class;
+  GList *combined_outputs;
   GfOutput *output;
 
-  manager = GF_MONITOR_MANAGER (skeleton);
-  manager_class = GF_MONITOR_MANAGER_GET_CLASS (skeleton);
+  manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
 
   if (serial != manager->serial)
     {
@@ -1310,15 +1256,20 @@ gf_monitor_manager_handle_change_backlight (GfDBusDisplayConfig   *skeleton,
       return TRUE;
     }
 
-  if (output_index >= g_list_length (manager->outputs))
+  combined_outputs = combine_gpu_lists (manager, gf_gpu_get_outputs);
+
+  if (output_index >= g_list_length (combined_outputs))
+
     {
+      g_list_free (combined_outputs);
       g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR,
                                              G_DBUS_ERROR_INVALID_ARGS,
                                              "Invalid output id");
       return TRUE;
     }
 
-  output = g_list_nth_data (manager->outputs, output_index);
+  output = g_list_nth_data (combined_outputs, output_index);
+  g_list_free (combined_outputs);
 
   if (value < 0 || value > 100)
     {
@@ -1349,10 +1300,11 @@ static gboolean
 gf_monitor_manager_handle_get_crtc_gamma (GfDBusDisplayConfig   *skeleton,
                                           GDBusMethodInvocation *invocation,
                                           guint                  serial,
-                                          guint                  crtc_id)
+                                          guint                  crtc_id,
+                                          GfMonitorManager      *manager)
 {
-  GfMonitorManager *manager;
   GfMonitorManagerClass *manager_class;
+  GList *combined_crtcs;
   GfCrtc *crtc;
   gsize size;
   gushort *red;
@@ -1365,8 +1317,7 @@ gf_monitor_manager_handle_get_crtc_gamma (GfDBusDisplayConfig   *skeleton,
   GVariant *green_v;
   GVariant *blue_v;
 
-  manager = GF_MONITOR_MANAGER (skeleton);
-  manager_class = GF_MONITOR_MANAGER_GET_CLASS (skeleton);
+  manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
 
   if (serial != manager->serial)
     {
@@ -1376,15 +1327,19 @@ gf_monitor_manager_handle_get_crtc_gamma (GfDBusDisplayConfig   *skeleton,
       return TRUE;
     }
 
-  if (crtc_id >= g_list_length (manager->crtcs))
+  combined_crtcs = combine_gpu_lists (manager, gf_gpu_get_crtcs);
+
+  if (crtc_id >= g_list_length (combined_crtcs))
     {
+      g_list_free (combined_crtcs);
       g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR,
                                              G_DBUS_ERROR_INVALID_ARGS,
                                              "Invalid crtc id");
       return TRUE;
     }
 
-  crtc = g_list_nth_data (manager->crtcs, crtc_id);
+  crtc = g_list_nth_data (combined_crtcs, crtc_id);
+  g_list_free (combined_crtcs);
 
   if (manager_class->get_crtc_gamma)
     {
@@ -1421,10 +1376,11 @@ gf_monitor_manager_handle_set_crtc_gamma (GfDBusDisplayConfig   *skeleton,
                                           guint                  crtc_id,
                                           GVariant              *red_v,
                                           GVariant              *green_v,
-                                          GVariant              *blue_v)
+                                          GVariant              *blue_v,
+                                          GfMonitorManager      *manager)
 {
-  GfMonitorManager *manager;
   GfMonitorManagerClass *manager_class;
+  GList *combined_crtcs;
   GfCrtc *crtc;
   GBytes *red_bytes;
   GBytes *green_bytes;
@@ -1434,8 +1390,7 @@ gf_monitor_manager_handle_set_crtc_gamma (GfDBusDisplayConfig   *skeleton,
   gushort *green;
   gushort *blue;
 
-  manager = GF_MONITOR_MANAGER (skeleton);
-  manager_class = GF_MONITOR_MANAGER_GET_CLASS (skeleton);
+  manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
 
   if (serial != manager->serial)
     {
@@ -1445,15 +1400,19 @@ gf_monitor_manager_handle_set_crtc_gamma (GfDBusDisplayConfig   *skeleton,
       return TRUE;
     }
 
-  if (crtc_id >= g_list_length (manager->crtcs))
+  combined_crtcs = combine_gpu_lists (manager, gf_gpu_get_crtcs);
+
+  if (crtc_id >= g_list_length (combined_crtcs))
     {
+      g_list_free (combined_crtcs);
       g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR,
                                              G_DBUS_ERROR_INVALID_ARGS,
                                              "Invalid crtc id");
       return TRUE;
     }
 
-  crtc = g_list_nth_data (manager->crtcs, crtc_id);
+  crtc = g_list_nth_data (combined_crtcs, crtc_id);
+  g_list_free (combined_crtcs);
 
   red_bytes = g_variant_get_data_as_bytes (red_v);
   green_bytes = g_variant_get_data_as_bytes (green_v);
@@ -1488,9 +1447,9 @@ gf_monitor_manager_handle_set_crtc_gamma (GfDBusDisplayConfig   *skeleton,
 
 static gboolean
 gf_monitor_manager_handle_get_current_state (GfDBusDisplayConfig   *skeleton,
-                                             GDBusMethodInvocation *invocation)
+                                             GDBusMethodInvocation *invocation,
+                                             GfMonitorManager      *manager)
 {
-  GfMonitorManager *manager;
   GfMonitorManagerPrivate *priv;
   GfSettings *settings;
   GVariantBuilder monitors_builder;
@@ -1502,7 +1461,6 @@ gf_monitor_manager_handle_get_current_state (GfDBusDisplayConfig   *skeleton,
   gint max_screen_width;
   gint max_screen_height;
 
-  manager = GF_MONITOR_MANAGER (skeleton);
   priv = gf_monitor_manager_get_instance_private (manager);
   settings = gf_backend_get_settings (priv->backend);
 
@@ -1520,8 +1478,7 @@ gf_monitor_manager_handle_get_current_state (GfDBusDisplayConfig   *skeleton,
       GVariantBuilder monitor_properties_builder;
       GList *k;
       gboolean is_builtin;
-      GfOutput *main_output;
-      gchar *display_name;
+      const char *display_name;
       gint i;
 
       current_mode = gf_monitor_get_current_mode (monitor);
@@ -1607,11 +1564,10 @@ gf_monitor_manager_handle_get_current_state (GfDBusDisplayConfig   *skeleton,
                              "is-builtin",
                              g_variant_new_boolean (is_builtin));
 
-      main_output = gf_monitor_get_main_output (monitor);
-      display_name = make_display_name (manager, main_output);
+      display_name = gf_monitor_get_display_name (monitor);
       g_variant_builder_add (&monitor_properties_builder, "{sv}",
                              "display-name",
-                             g_variant_new_take_string (display_name));
+                             g_variant_new_string (display_name));
 
       g_variant_builder_add (&monitors_builder, MONITOR_FORMAT,
                              monitor_spec->connector,
@@ -1725,9 +1681,9 @@ gf_monitor_manager_handle_apply_monitors_config (GfDBusDisplayConfig   *skeleton
                                                  guint                  serial,
                                                  guint                  method,
                                                  GVariant              *logical_monitor_configs_variant,
-                                                 GVariant              *properties_variant)
+                                                 GVariant              *properties_variant,
+                                                 GfMonitorManager      *manager)
 {
-  GfMonitorManager *manager;
   GfMonitorManagerCapability capabilities;
   GVariant *layout_mode_variant;
   GfLogicalMonitorLayoutMode layout_mode;
@@ -1736,8 +1692,6 @@ gf_monitor_manager_handle_apply_monitors_config (GfDBusDisplayConfig   *skeleton
   GError *error;
   GfMonitorsConfig *config;
 
-  manager = GF_MONITOR_MANAGER (skeleton);
-
   if (serial != manager->serial)
     {
       g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR,
@@ -1867,14 +1821,26 @@ gf_monitor_manager_handle_apply_monitors_config (GfDBusDisplayConfig   *skeleton
 }
 
 static void
-gf_monitor_manager_display_config_init (GfDBusDisplayConfigIface *iface)
+monitor_manager_setup_dbus_config_handlers (GfMonitorManager *manager)
 {
-  iface->handle_get_resources = gf_monitor_manager_handle_get_resources;
-  iface->handle_change_backlight = gf_monitor_manager_handle_change_backlight;
-  iface->handle_get_crtc_gamma = gf_monitor_manager_handle_get_crtc_gamma;
-  iface->handle_set_crtc_gamma = gf_monitor_manager_handle_set_crtc_gamma;
-  iface->handle_get_current_state = gf_monitor_manager_handle_get_current_state;
-  iface->handle_apply_monitors_config = gf_monitor_manager_handle_apply_monitors_config;
+  g_signal_connect_object (manager->display_config, "handle-get-resources",
+                           G_CALLBACK (gf_monitor_manager_handle_get_resources),
+                           manager, 0);
+  g_signal_connect_object (manager->display_config, "handle-change-backlight",
+                           G_CALLBACK (gf_monitor_manager_handle_change_backlight),
+                           manager, 0);
+  g_signal_connect_object (manager->display_config, "handle-get-crtc-gamma",
+                           G_CALLBACK (gf_monitor_manager_handle_get_crtc_gamma),
+                           manager, 0);
+  g_signal_connect_object (manager->display_config, "handle-set-crtc-gamma",
+                           G_CALLBACK (gf_monitor_manager_handle_set_crtc_gamma),
+                           manager, 0);
+  g_signal_connect_object (manager->display_config, "handle-get-current-state",
+                           G_CALLBACK (gf_monitor_manager_handle_get_current_state),
+                           manager, 0);
+  g_signal_connect_object (manager->display_config, "handle-apply-monitors-config",
+                           G_CALLBACK (gf_monitor_manager_handle_apply_monitors_config),
+                           manager, 0);
 }
 
 static void
@@ -1886,7 +1852,7 @@ bus_acquired_cb (GDBusConnection *connection,
   GDBusInterfaceSkeleton *skeleton;
 
   manager = GF_MONITOR_MANAGER (user_data);
-  skeleton = G_DBUS_INTERFACE_SKELETON (manager);
+  skeleton = G_DBUS_INTERFACE_SKELETON (manager->display_config);
 
   g_dbus_interface_skeleton_export (skeleton, connection,
                                     "/org/gnome/Mutter/DisplayConfig",
@@ -1914,20 +1880,13 @@ gf_monitor_manager_real_read_edid (GfMonitorManager *manager,
   return NULL;
 }
 
-static gchar *
-gf_monitor_manager_real_get_edid_file (GfMonitorManager *manager,
-                                       GfOutput         *output)
-{
-  return NULL;
-}
-
 static gboolean
 gf_monitor_manager_real_is_lid_closed (GfMonitorManager *manager)
 {
   if (!manager->up_client)
     return FALSE;
 
-  return up_client_get_lid_is_closed (manager->up_client);
+  return manager->lid_is_closed;
 }
 
 static void
@@ -1935,8 +1894,16 @@ lid_is_closed_changed (UpClient   *client,
                        GParamSpec *pspec,
                        gpointer    user_data)
 {
-  GfMonitorManager *manager = user_data;
+  GfMonitorManager *manager;
+  gboolean lid_is_closed;
 
+  manager = user_data;
+  lid_is_closed = up_client_get_lid_is_closed (manager->up_client);
+
+  if (lid_is_closed == manager->lid_is_closed)
+    return;
+
+  manager->lid_is_closed = lid_is_closed;
   gf_monitor_manager_ensure_configured (manager);
 }
 
@@ -1954,16 +1921,25 @@ gf_monitor_manager_constructed (GObject *object)
 
   G_OBJECT_CLASS (gf_monitor_manager_parent_class)->constructed (object);
 
+  manager->display_config = gf_dbus_display_config_skeleton_new ();
+  monitor_manager_setup_dbus_config_handlers (manager);
+
   if (manager_class->is_lid_closed == gf_monitor_manager_real_is_lid_closed)
     {
       manager->up_client = up_client_new ();
 
-      g_signal_connect_object (manager->up_client, "notify::lid-is-closed",
-                               G_CALLBACK (lid_is_closed_changed), manager, 0);
+      if (manager->up_client)
+        {
+          g_signal_connect_object (manager->up_client, "notify::lid-is-closed",
+                                   G_CALLBACK (lid_is_closed_changed), manager, 0);
+
+          manager->lid_is_closed = up_client_get_lid_is_closed (manager->up_client);
+        }
     }
 
-  g_signal_connect_object (manager, "notify::power-save-mode",
-                           G_CALLBACK (power_save_mode_changed), manager, 0);
+  g_signal_connect_object (manager->display_config, "notify::power-save-mode",
+                           G_CALLBACK (power_save_mode_changed), manager,
+                           G_CONNECT_SWAPPED);
 
   orientation_manager = gf_backend_get_orientation_manager (priv->backend);
   g_signal_connect_object (orientation_manager, "orientation-changed",
@@ -1971,11 +1947,6 @@ gf_monitor_manager_constructed (GObject *object)
 
   manager->current_switch_config = GF_MONITOR_SWITCH_CONFIG_UNKNOWN;
 
-  manager->config_manager = gf_monitor_config_manager_new (manager);
-
-  gf_monitor_manager_read_current_state (manager);
-  manager_class->ensure_initial_config (manager);
-
   priv->bus_name_id = g_bus_own_name (G_BUS_TYPE_SESSION,
                                       "org.gnome.Mutter.DisplayConfig",
                                       G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT |
@@ -1985,8 +1956,6 @@ gf_monitor_manager_constructed (GObject *object)
                                       name_lost_cb,
                                       g_object_ref (manager),
                                       g_object_unref);
-
-  priv->in_init = FALSE;
 }
 
 static void
@@ -2004,6 +1973,7 @@ gf_monitor_manager_dispose (GObject *object)
       priv->bus_name_id = 0;
     }
 
+  g_clear_object (&manager->display_config);
   g_clear_object (&manager->config_manager);
   g_clear_object (&manager->up_client);
 
@@ -2019,10 +1989,7 @@ gf_monitor_manager_finalize (GObject *object)
 
   manager = GF_MONITOR_MANAGER (object);
 
-  g_list_free_full (manager->outputs, g_object_unref);
-  g_list_free_full (manager->modes, g_object_unref);
-  g_list_free_full (manager->crtcs, g_object_unref);
-
+  g_list_free_full (manager->gpus, g_object_unref);
   g_list_free_full (manager->logical_monitors, g_object_unref);
 
   G_OBJECT_CLASS (gf_monitor_manager_parent_class)->finalize (object);
@@ -2095,10 +2062,28 @@ gf_monitor_manager_install_properties (GObjectClass *object_class)
 static void
 gf_monitor_manager_install_signals (GObjectClass *object_class)
 {
+  manager_signals[MONITORS_CHANGED] =
+    g_signal_new ("monitors-changed",
+                  G_TYPE_FROM_CLASS (object_class),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
+
+  manager_signals[POWER_SAVE_MODE_CHANGED] =
+    g_signal_new ("power-save-mode-changed",
+                  G_TYPE_FROM_CLASS (object_class),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
+
   manager_signals[CONFIRM_DISPLAY_CHANGE] =
     g_signal_new ("confirm-display-change",
-                  G_TYPE_FROM_CLASS (object_class), G_SIGNAL_RUN_LAST,
-                  0, NULL, NULL, NULL, G_TYPE_NONE, 0);
+                  G_TYPE_FROM_CLASS (object_class),
+                  G_SIGNAL_RUN_LAST,
+                  0, NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
 }
 
 static void
@@ -2114,7 +2099,6 @@ gf_monitor_manager_class_init (GfMonitorManagerClass *manager_class)
   object_class->get_property = gf_monitor_manager_get_property;
   object_class->set_property = gf_monitor_manager_set_property;
 
-  manager_class->get_edid_file = gf_monitor_manager_real_get_edid_file;
   manager_class->read_edid = gf_monitor_manager_real_read_edid;
   manager_class->is_lid_closed = gf_monitor_manager_real_is_lid_closed;
 
@@ -2125,11 +2109,6 @@ gf_monitor_manager_class_init (GfMonitorManagerClass *manager_class)
 static void
 gf_monitor_manager_init (GfMonitorManager *manager)
 {
-  GfMonitorManagerPrivate *priv;
-
-  priv = gf_monitor_manager_get_instance_private (manager);
-
-  priv->in_init = TRUE;
 }
 
 GfBackend *
@@ -2142,6 +2121,25 @@ gf_monitor_manager_get_backend (GfMonitorManager *manager)
   return priv->backend;
 }
 
+void
+gf_monitor_manager_setup (GfMonitorManager *manager)
+{
+  GfMonitorManagerPrivate *priv;
+  GfMonitorManagerClass *manager_class;
+
+  priv = gf_monitor_manager_get_instance_private (manager);
+  manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
+
+  priv->in_init = TRUE;
+
+  manager->config_manager = gf_monitor_config_manager_new (manager);
+
+  gf_monitor_manager_read_current_state (manager);
+  manager_class->ensure_initial_config (manager);
+
+  priv->in_init = FALSE;
+}
+
 void
 gf_monitor_manager_rebuild_derived (GfMonitorManager *manager,
                                     GfMonitorsConfig *config)
@@ -2211,16 +2209,17 @@ gf_monitor_manager_get_primary_logical_monitor (GfMonitorManager *manager)
   return manager->primary_logical_monitor;
 }
 
-GList *
-gf_monitor_manager_get_outputs (GfMonitorManager *manager)
+void
+gf_monitor_manager_add_gpu (GfMonitorManager *manager,
+                            GfGpu            *gpu)
 {
-  return manager->outputs;
+  manager->gpus = g_list_append (manager->gpus, gpu);
 }
 
 GList *
-gf_monitor_manager_get_crtcs (GfMonitorManager *manager)
+gf_monitor_manager_get_gpus (GfMonitorManager *manager)
 {
-  return manager->crtcs;
+  return manager->gpus;
 }
 
 gboolean
@@ -2228,11 +2227,11 @@ gf_monitor_manager_has_hotplug_mode_update (GfMonitorManager *manager)
 {
   GList *l;
 
-  for (l = manager->outputs; l; l = l->next)
+  for (l = manager->gpus; l; l = l->next)
     {
-      GfOutput *output = l->data;
+      GfGpu *gpu = l->data;
 
-      if (output->hotplug_mode_update)
+      if (gf_gpu_has_hotplug_mode_update (gpu))
         return TRUE;
     }
 
@@ -2242,26 +2241,23 @@ gf_monitor_manager_has_hotplug_mode_update (GfMonitorManager *manager)
 void
 gf_monitor_manager_read_current_state (GfMonitorManager *manager)
 {
-  GList *old_outputs;
-  GList *old_crtcs;
-  GList *old_modes;
-
-  /* Some implementations of read_current use the existing information
-   * we have available, so don't free the old configuration until after
-   * read_current finishes.
-   */
-  old_outputs = manager->outputs;
-  old_crtcs = manager->crtcs;
-  old_modes = manager->modes;
+  GList *l;
 
   manager->serial++;
-  GF_MONITOR_MANAGER_GET_CLASS (manager)->read_current (manager);
+
+  for (l = manager->gpus; l; l = l->next)
+    {
+      GfGpu *gpu = l->data;
+      GError *error = NULL;
+
+      if (!gf_gpu_read_current (gpu, &error))
+        {
+          g_warning ("Failed to read current KMS state: %s", error->message);
+          g_clear_error (&error);
+        }
+    }
 
   rebuild_monitors (manager);
-
-  g_list_free_full (old_outputs, g_object_unref);
-  g_list_free_full (old_modes, g_object_unref);
-  g_list_free_full (old_crtcs, g_object_unref);
 }
 
 void
@@ -2272,6 +2268,7 @@ gf_monitor_manager_on_hotplug (GfMonitorManager *manager)
 
 gboolean
 gf_monitor_manager_get_monitor_matrix (GfMonitorManager *manager,
+                                       GfMonitor        *monitor,
                                        GfLogicalMonitor *logical_monitor,
                                        gfloat            matrix[6])
 {
@@ -2281,7 +2278,9 @@ gf_monitor_manager_get_monitor_matrix (GfMonitorManager *manager,
   if (!calculate_viewport_matrix (manager, logical_monitor, viewport))
     return FALSE;
 
+  /* Get transform corrected for LCD panel-orientation. */
   transform = logical_monitor->transform;
+  transform = gf_monitor_logical_to_crtc_transform (monitor, transform);
   multiply_matrix (viewport, transform_matrices[transform], matrix);
 
   return TRUE;
@@ -2405,6 +2404,7 @@ gf_monitor_manager_ensure_configured (GfMonitorManager *manager)
     }
 
   config = gf_monitor_config_manager_create_linear (manager->config_manager);
+
   if (config)
     {
       if (!gf_monitor_manager_apply_monitors_config (manager, config,
@@ -2457,6 +2457,11 @@ void
 gf_monitor_manager_update_logical_state_derived (GfMonitorManager *manager,
                                                  GfMonitorsConfig *config)
 {
+  if (config)
+    manager->current_switch_config = gf_monitors_config_get_switch_config (config);
+  else
+    manager->current_switch_config = GF_MONITOR_SWITCH_CONFIG_UNKNOWN;
+
   manager->layout_mode = GF_LOGICAL_MONITOR_LAYOUT_MODE_PHYSICAL;
 
   gf_monitor_manager_rebuild_logical_monitors_derived (manager, config);
@@ -2562,6 +2567,16 @@ gf_monitor_manager_get_config_manager (GfMonitorManager *manager)
   return manager->config_manager;
 }
 
+char *
+gf_monitor_manager_get_vendor_name (GfMonitorManager *manager,
+                                    const char       *vendor)
+{
+  if (!manager->pnp_ids)
+    manager->pnp_ids = gnome_pnp_ids_new ();
+
+  return gnome_pnp_ids_get_pnp_id (manager->pnp_ids, vendor);
+}
+
 gint
 gf_monitor_manager_get_monitor_for_connector (GfMonitorManager *manager,
                                               const gchar      *connector)
@@ -2574,11 +2589,7 @@ gf_monitor_manager_get_monitor_for_connector (GfMonitorManager *manager,
 
       if (gf_monitor_is_active (monitor) &&
           g_str_equal (connector, gf_monitor_get_connector (monitor)))
-        {
-          GfOutput *main_output = gf_monitor_get_main_output (monitor);
-
-          return main_output->crtc->logical_monitor->number;
-        }
+        return gf_monitor_get_logical_monitor (monitor)->number;
     }
 
   return -1;
