@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2016 Red Hat
- * Copyright (C) 2017 Alberts Muktupāvels
+ * Copyright (C) 2017-2019 Alberts Muktupāvels
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -48,14 +48,15 @@ generate_modes (GfMonitorNormal *normal)
       GfCrtcMode *crtc_mode;
       GfMonitorMode *mode;
       gboolean replace;
+      GfCrtc *crtc;
 
       crtc_mode = output->modes[i];
 
       mode = g_new0 (GfMonitorMode, 1);
-      mode->spec.width = crtc_mode->width;
-      mode->spec.height = crtc_mode->height;
-      mode->spec.refresh_rate = crtc_mode->refresh_rate;
-      mode->spec.flags = crtc_mode->flags & HANDLED_CRTC_MODE_FLAGS;
+      mode->spec = gf_monitor_create_spec (monitor,
+                                           crtc_mode->width,
+                                           crtc_mode->height,
+                                           crtc_mode);
 
       mode->id = gf_monitor_mode_spec_generate_id (&mode->spec);
 
@@ -82,7 +83,8 @@ generate_modes (GfMonitorNormal *normal)
       if (crtc_mode == output->preferred_mode)
         gf_monitor_set_preferred_mode (monitor, mode);
 
-      if (output->crtc && crtc_mode == output->crtc->current_mode)
+      crtc = gf_output_get_assigned_crtc (output);
+      if (crtc && crtc_mode == crtc->current_mode)
         gf_monitor_set_current_mode (monitor, mode);
     }
 }
@@ -102,14 +104,16 @@ gf_monitor_normal_derive_layout (GfMonitor   *monitor,
                                  GfRectangle *layout)
 {
   GfOutput *output;
+  GfCrtc *crtc;
 
   output = gf_monitor_get_main_output (monitor);
+  crtc = gf_output_get_assigned_crtc (output);
 
   *layout = (GfRectangle) {
-    .x = output->crtc->rect.x,
-    .y = output->crtc->rect.y,
-    .width = output->crtc->rect.width,
-    .height = output->crtc->rect.height
+    .x = crtc->rect.x,
+    .y = crtc->rect.y,
+    .width = crtc->rect.width,
+    .height = crtc->rect.height
   };
 }
 
@@ -161,14 +165,14 @@ gf_monitor_normal_init (GfMonitorNormal *normal)
 }
 
 GfMonitorNormal *
-gf_monitor_normal_new (GfMonitorManager *monitor_manager,
-                       GfOutput         *output)
+gf_monitor_normal_new (GfGpu    *gpu,
+                       GfOutput *output)
 {
   GfMonitorNormal *normal;
   GfMonitor *monitor;
 
   normal = g_object_new (GF_TYPE_MONITOR_NORMAL,
-                         "monitor-manager", monitor_manager,
+                         "gpu", gpu,
                          NULL);
 
   monitor = GF_MONITOR (normal);
@@ -178,5 +182,7 @@ gf_monitor_normal_new (GfMonitorManager *monitor_manager,
   gf_monitor_generate_spec (monitor);
   generate_modes (normal);
 
+  gf_monitor_make_display_name (monitor);
+
   return normal;
 }
