@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2016 Red Hat
- * Copyright (C) 2017 Alberts Muktupāvels
+ * Copyright (C) 2017-2019 Alberts Muktupāvels
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -138,10 +138,10 @@ create_tiled_monitor_mode (GfMonitorTiled *tiled,
 
   calculate_tiled_size (monitor, &width, &height);
 
-  mode->parent.spec.width = width;
-  mode->parent.spec.height = height;
-  mode->parent.spec.refresh_rate = reference_crtc_mode->refresh_rate;
-  mode->parent.spec.flags = reference_crtc_mode->flags & HANDLED_CRTC_MODE_FLAGS;
+  mode->parent.spec = gf_monitor_create_spec (monitor,
+                                              width,
+                                              height,
+                                              reference_crtc_mode);
 
   mode->parent.id = gf_monitor_mode_spec_generate_id (&mode->parent.spec);
   mode->parent.crtc_modes = g_new0 (GfMonitorCrtcMode, g_list_length (outputs));
@@ -192,10 +192,10 @@ create_untiled_monitor_mode (GfMonitorTiled *tiled,
   mode = g_new0 (GfMonitorModeTiled, 1);
   mode->is_tiled = FALSE;
 
-  mode->parent.spec.width = crtc_mode->width;
-  mode->parent.spec.height = crtc_mode->height;
-  mode->parent.spec.refresh_rate = crtc_mode->refresh_rate;
-  mode->parent.spec.flags = crtc_mode->flags & HANDLED_CRTC_MODE_FLAGS;
+  mode->parent.spec = gf_monitor_create_spec (monitor,
+                                              crtc_mode->width,
+                                              crtc_mode->height,
+                                              crtc_mode);
 
   mode->parent.id = gf_monitor_mode_spec_generate_id (&mode->parent.spec);
   mode->parent.crtc_modes = g_new0 (GfMonitorCrtcMode, g_list_length (outputs));
@@ -494,8 +494,8 @@ find_untiled_output (GfMonitorTiled *tiled)
 }
 
 static void
-add_tiled_monitor_outputs (GfMonitorManager *monitor_manager,
-                           GfMonitorTiled   *tiled)
+add_tiled_monitor_outputs (GfGpu          *gpu,
+                           GfMonitorTiled *tiled)
 {
   GfMonitor *monitor;
   GList *outputs;
@@ -503,7 +503,7 @@ add_tiled_monitor_outputs (GfMonitorManager *monitor_manager,
 
   monitor = GF_MONITOR (tiled);
 
-  outputs = monitor_manager->outputs;
+  outputs = gf_gpu_get_outputs (gpu);
   for (l = outputs; l; l = l->next)
     {
       GfOutput *output;
@@ -593,11 +593,13 @@ static void
 gf_monitor_tiled_finalize (GObject *object)
 {
   GfMonitor *monitor;
+  GfGpu *gpu;
   GfMonitorManager *monitor_manager;
 
   monitor = GF_MONITOR (object);
 
-  monitor_manager = gf_monitor_get_monitor_manager (monitor);
+  gpu = gf_monitor_get_gpu (monitor);
+  monitor_manager = gf_gpu_get_monitor_manager (gpu);
   gf_monitor_manager_tiled_monitor_removed (monitor_manager, monitor);
 
   G_OBJECT_CLASS (gf_monitor_tiled_parent_class)->finalize (object);
@@ -633,15 +635,19 @@ gf_monitor_tiled_derive_layout (GfMonitor   *monitor,
 
   for (l = outputs; l; l = l->next)
     {
-      GfOutput *output = l->data;
+      GfOutput *output;
+      GfCrtc *crtc;
 
-      if (!output->crtc)
+      output = l->data;
+      crtc = gf_output_get_assigned_crtc (output);
+
+      if (!crtc)
         continue;
 
-      min_x = MIN (output->crtc->rect.x, min_x);
-      min_y = MIN (output->crtc->rect.y, min_y);
-      max_x = MAX (output->crtc->rect.x + output->crtc->rect.width, max_x);
-      max_y = MAX (output->crtc->rect.y + output->crtc->rect.height, max_y);
+      min_x = MIN (crtc->rect.x, min_x);
+      min_y = MIN (crtc->rect.y, min_y);
+      max_x = MAX (crtc->rect.x + crtc->rect.width, max_x);
+      max_y = MAX (crtc->rect.y + crtc->rect.height, max_y);
     }
 
   *layout = (GfRectangle) {
@@ -706,14 +712,15 @@ gf_monitor_tiled_init (GfMonitorTiled *tiled)
 }
 
 GfMonitorTiled *
-gf_monitor_tiled_new (GfMonitorManager *monitor_manager,
-                      GfOutput         *output)
+gf_monitor_tiled_new (GfGpu    *gpu,
+                      GfOutput *output)
 {
   GfMonitorTiled *tiled;
   GfMonitor *monitor;
+  GfMonitorManager *monitor_manager;
 
   tiled = g_object_new (GF_TYPE_MONITOR_TILED,
-                        "monitor-manager", monitor_manager,
+                        "gpu", gpu,
                         NULL);
 
   monitor = GF_MONITOR (tiled);
@@ -722,15 +729,18 @@ gf_monitor_tiled_new (GfMonitorManager *monitor_manager,
   gf_monitor_set_winsys_id (monitor, output->winsys_id);
 
   tiled->origin_output = output;
-  add_tiled_monitor_outputs (monitor_manager, tiled);
+  add_tiled_monitor_outputs (gpu, tiled);
 
   tiled->main_output = find_untiled_output (tiled);
 
   gf_monitor_generate_spec (monitor);
 
+  monitor_manager = gf_gpu_get_monitor_manager (gpu);
   gf_monitor_manager_tiled_monitor_added (monitor_manager, monitor);
   generate_modes (tiled);
 
+  gf_monitor_make_display_name (monitor);
+
   return tiled;
 }
 
