@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2016 Red Hat
- * Copyright (C) 2017 Alberts Muktupāvels
+ * Copyright (C) 2017-2019 Alberts Muktupāvels
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,10 +21,12 @@
 
 #include "config.h"
 
+#include <glib/gi18n-lib.h>
 #include <math.h>
 #include <string.h>
 
 #include "gf-crtc-private.h"
+#include "gf-gpu-private.h"
 #include "gf-monitor-manager-private.h"
 #include "gf-monitor-private.h"
 #include "gf-monitor-spec-private.h"
@@ -52,16 +54,16 @@
 
 typedef struct
 {
-  GfMonitorManager *monitor_manager;
+  GfGpu         *gpu;
 
-  GList            *outputs;
-  GList            *modes;
-  GHashTable       *mode_ids;
+  GList         *outputs;
+  GList         *modes;
+  GHashTable    *mode_ids;
 
-  GfMonitorMode    *preferred_mode;
-  GfMonitorMode    *current_mode;
+  GfMonitorMode *preferred_mode;
+  GfMonitorMode *current_mode;
 
-  GfMonitorSpec    *spec;
+  GfMonitorSpec *spec;
 
   /*
    * The primary or first output for this monitor, 0 if we can't figure out.
@@ -72,14 +74,16 @@ typedef struct
    * (it's an attempt to keep windows on the same monitor, and preferably on
    * the primary one).
    */
-  glong             winsys_id;
+  glong          winsys_id;
+
+  char          *display_name;
 } GfMonitorPrivate;
 
 enum
 {
   PROP_0,
 
-  PROP_MONITOR_MANAGER,
+  PROP_GPU,
 
   LAST_PROP
 };
@@ -88,6 +92,124 @@ static GParamSpec *monitor_properties[LAST_PROP] = { NULL };
 
 G_DEFINE_TYPE_WITH_PRIVATE (GfMonitor, gf_monitor, G_TYPE_OBJECT)
 
+static const gdouble known_diagonals[] =
+  {
+    12.1,
+    13.3,
+    15.6
+  };
+
+static gchar *
+diagonal_to_str (gdouble d)
+{
+  guint i;
+
+  for (i = 0; i < G_N_ELEMENTS (known_diagonals); i++)
+    {
+      gdouble delta;
+
+      delta = fabs(known_diagonals[i] - d);
+
+      if (delta < 0.1)
+        return g_strdup_printf ("%0.1lf\"", known_diagonals[i]);
+    }
+
+  return g_strdup_printf ("%d\"", (int) (d + 0.5));
+}
+
+static gchar *
+make_display_name (GfMonitor        *monitor,
+                   GfMonitorManager *manager)
+{
+  gchar *inches;
+  gchar *vendor_name;
+  const char *vendor;
+  const char *product_name;
+  int width_mm;
+  int height_mm;
+
+  if (gf_monitor_is_laptop_panel (monitor))
+    return g_strdup (_("Built-in display"));
+
+  inches = NULL;
+  vendor_name = NULL;
+  vendor = gf_monitor_get_vendor (monitor);
+  product_name = NULL;
+
+  gf_monitor_get_physical_dimensions (monitor, &width_mm, &height_mm);
+
+  if (width_mm > 0 && height_mm > 0)
+    {
+      if (!gf_monitor_has_aspect_as_size (monitor))
+        {
+          double d;
+
+          d = sqrt (width_mm * width_mm + height_mm * height_mm);
+          inches = diagonal_to_str (d / 25.4);
+        }
+      else
+        {
+          product_name = gf_monitor_get_product (monitor);
+        }
+    }
+
+  if (g_strcmp0 (vendor, "unknown") != 0)
+    {
+      vendor_name = gf_monitor_manager_get_vendor_name (manager, vendor);
+
+      if (!vendor_name)
+        vendor_name = g_strdup (vendor);
+    }
+  else
+    {
+      if (inches != NULL)
+        vendor_name = g_strdup (_("Unknown"));
+      else
+        vendor_name = g_strdup (_("Unknown Display"));
+    }
+
+  if (inches != NULL)
+    {
+      gchar *display_name;
+
+      display_name = g_strdup_printf (C_("This is a monitor vendor name, followed by a "
+                                         "size in inches, like 'Dell 15\"'",
+                                         "%s %s"), vendor_name, inches);
+
+      g_free (vendor_name);
+      g_free (inches);
+
+      return display_name;
+    }
+  else if (product_name != NULL)
+    {
+      gchar *display_name;
+
+      display_name =  g_strdup_printf (C_("This is a monitor vendor name followed by "
+                                          "product/model name where size in inches "
+                                          "could not be calculated, e.g. Dell U2414H",
+                                          "%s %s"), vendor_name, product_name);
+
+      g_free (vendor_name);
+
+      return display_name;
+    }
+
+  return vendor_name;
+}
+
+static gboolean
+is_current_mode_known (GfMonitor *monitor)
+{
+  GfOutput *output;
+  GfCrtc *crtc;
+
+  output = gf_monitor_get_main_output (monitor);
+  crtc = gf_output_get_assigned_crtc (output);
+
+  return gf_monitor_is_active (monitor) == (crtc && crtc->current_mode);
+}
+
 static gboolean
 gf_monitor_mode_spec_equals (GfMonitorModeSpec *spec,
                              GfMonitorModeSpec *other_spec)
@@ -152,10 +274,7 @@ calculate_scale (GfMonitor     *monitor,
   /* Somebody encoded the aspect ratio (16/9 or 16/10) instead of the
    * physical size.
    */
-  if ((width_mm == 160 && height_mm == 90) ||
-      (width_mm == 160 && height_mm == 100) ||
-      (width_mm == 16 && height_mm == 9) ||
-      (width_mm == 16 && height_mm == 10))
+  if (gf_monitor_has_aspect_as_size (monitor))
     goto out;
 
   if (width_mm > 0 && height_mm > 0)
@@ -247,6 +366,24 @@ out:
   return best_scale;
 }
 
+static void
+gf_monitor_dispose (GObject *object)
+{
+  GfMonitor *monitor;
+  GfMonitorPrivate *priv;
+
+  monitor = GF_MONITOR (object);
+  priv = gf_monitor_get_instance_private (monitor);
+
+  if (priv->outputs)
+    {
+      g_list_free_full (priv->outputs, g_object_unref);
+      priv->outputs = NULL;
+    }
+
+  G_OBJECT_CLASS (gf_monitor_parent_class)->dispose (object);
+}
+
 static void
 gf_monitor_finalize (GObject *object)
 {
@@ -258,8 +395,8 @@ gf_monitor_finalize (GObject *object)
 
   g_hash_table_destroy (priv->mode_ids);
   g_list_free_full (priv->modes, (GDestroyNotify) gf_monitor_mode_free);
-  g_clear_pointer (&priv->outputs, g_list_free);
   gf_monitor_spec_free (priv->spec);
+  g_free (priv->display_name);
 
   G_OBJECT_CLASS (gf_monitor_parent_class)->finalize (object);
 }
@@ -278,8 +415,8 @@ gf_monitor_get_property (GObject    *object,
 
   switch (property_id)
     {
-      case PROP_MONITOR_MANAGER:
-        g_value_set_object (value, priv->monitor_manager);
+      case PROP_GPU:
+        g_value_set_object (value, priv->gpu);
         break;
 
       default:
@@ -302,8 +439,8 @@ gf_monitor_set_property (GObject      *object,
 
   switch (property_id)
     {
-      case PROP_MONITOR_MANAGER:
-        priv->monitor_manager = g_value_get_object (value);
+      case PROP_GPU:
+        priv->gpu = g_value_get_object (value);
         break;
 
       default:
@@ -315,11 +452,11 @@ gf_monitor_set_property (GObject      *object,
 static void
 gf_monitor_install_properties (GObjectClass *object_class)
 {
-  monitor_properties[PROP_MONITOR_MANAGER] =
-    g_param_spec_object ("monitor-manager",
-                         "GfMonitorManager",
-                         "GfMonitorManager",
-                         GF_TYPE_MONITOR_MANAGER,
+  monitor_properties[PROP_GPU] =
+    g_param_spec_object ("gpu",
+                         "GfGpu",
+                         "GfGpu",
+                         GF_TYPE_GPU,
                          G_PARAM_READWRITE |
                          G_PARAM_STATIC_STRINGS |
                          G_PARAM_CONSTRUCT_ONLY);
@@ -335,6 +472,7 @@ gf_monitor_class_init (GfMonitorClass *monitor_class)
 
   object_class = G_OBJECT_CLASS (monitor_class);
 
+  object_class->dispose = gf_monitor_dispose;
   object_class->finalize = gf_monitor_finalize;
   object_class->get_property = gf_monitor_get_property;
   object_class->set_property = gf_monitor_set_property;
@@ -352,14 +490,38 @@ gf_monitor_init (GfMonitor *monitor)
   priv->mode_ids = g_hash_table_new (g_str_hash, g_str_equal);
 }
 
-GfMonitorManager *
-gf_monitor_get_monitor_manager (GfMonitor *monitor)
+GfGpu *
+gf_monitor_get_gpu (GfMonitor *monitor)
 {
   GfMonitorPrivate *priv;
 
   priv = gf_monitor_get_instance_private (monitor);
 
-  return priv->monitor_manager;
+  return priv->gpu;
+}
+
+void
+gf_monitor_make_display_name (GfMonitor *monitor)
+{
+  GfMonitorPrivate *priv;
+  GfMonitorManager *manager;
+
+  priv = gf_monitor_get_instance_private (monitor);
+
+  manager = gf_gpu_get_monitor_manager (priv->gpu);
+
+  g_free (priv->display_name);
+  priv->display_name = make_display_name (monitor, manager);
+}
+
+const char *
+gf_monitor_get_display_name (GfMonitor *monitor)
+{
+  GfMonitorPrivate *priv;
+
+  priv = gf_monitor_get_instance_private (monitor);
+
+  return priv->display_name;
 }
 
 gboolean
@@ -374,14 +536,18 @@ gf_monitor_is_mode_assigned (GfMonitor     *monitor,
 
   for (l = priv->outputs, i = 0; l; l = l->next, i++)
     {
-      GfOutput *output = l->data;
-      GfMonitorCrtcMode *monitor_crtc_mode = &mode->crtc_modes[i];
+      GfOutput *output;
+      GfMonitorCrtcMode *monitor_crtc_mode;
+      GfCrtc *crtc;
+
+      output = l->data;
+      monitor_crtc_mode = &mode->crtc_modes[i];
+      crtc = gf_output_get_assigned_crtc (output);
 
       if (monitor_crtc_mode->crtc_mode &&
-          (!output->crtc ||
-           output->crtc->current_mode != monitor_crtc_mode->crtc_mode))
+          (!crtc || crtc->current_mode != monitor_crtc_mode->crtc_mode))
         return FALSE;
-      else if (!monitor_crtc_mode->crtc_mode && output->crtc)
+      else if (!monitor_crtc_mode->crtc_mode && crtc)
         return FALSE;
     }
 
@@ -396,7 +562,7 @@ gf_monitor_append_output (GfMonitor *monitor,
 
   priv = gf_monitor_get_instance_private (monitor);
 
-  priv->outputs = g_list_append (priv->outputs, output);
+  priv->outputs = g_list_append (priv->outputs, g_object_ref (output));
 }
 
 void
@@ -421,6 +587,34 @@ gf_monitor_set_preferred_mode (GfMonitor     *monitor,
   priv->preferred_mode = mode;
 }
 
+GfMonitorModeSpec
+gf_monitor_create_spec (GfMonitor  *monitor,
+                        int         width,
+                        int         height,
+                        GfCrtcMode *crtc_mode)
+{
+  GfOutput *output;
+  GfMonitorModeSpec spec;
+
+  output = gf_monitor_get_main_output (monitor);
+
+  if (gf_monitor_transform_is_rotated (output->panel_orientation_transform))
+    {
+      int temp;
+
+      temp = width;
+      width = height;
+      height = temp;
+    }
+
+  spec.width = width;
+  spec.height = height;
+  spec.refresh_rate = crtc_mode->refresh_rate;
+  spec.flags = crtc_mode->flags & HANDLED_CRTC_MODE_FLAGS;
+
+  return spec;
+}
+
 void
 gf_monitor_generate_spec (GfMonitor *monitor)
 {
@@ -500,11 +694,11 @@ gf_monitor_get_spec (GfMonitor *monitor)
 gboolean
 gf_monitor_is_active (GfMonitor *monitor)
 {
-  GfOutput *output;
+  GfMonitorPrivate *priv;
 
-  output = gf_monitor_get_main_output (monitor);
+  priv = gf_monitor_get_instance_private (monitor);
 
-  return output->crtc && output->crtc->current_mode;
+  return !!priv->current_mode;
 }
 
 GfOutput *
@@ -677,6 +871,41 @@ gf_monitor_get_connector_type (GfMonitor *monitor)
   return output->connector_type;
 }
 
+GfMonitorTransform
+gf_monitor_logical_to_crtc_transform (GfMonitor          *monitor,
+                                      GfMonitorTransform  transform)
+{
+  GfOutput *output;
+  GfMonitorTransform new_transform;
+
+  output = gf_monitor_get_main_output (monitor);
+  new_transform = (transform + output->panel_orientation_transform) %
+                  GF_MONITOR_TRANSFORM_FLIPPED;
+
+  if (gf_monitor_transform_is_flipped (transform))
+    new_transform += GF_MONITOR_TRANSFORM_FLIPPED;
+
+  return new_transform;
+}
+
+GfMonitorTransform
+gf_monitor_crtc_to_logical_transform (GfMonitor          *monitor,
+                                      GfMonitorTransform  transform)
+{
+  GfOutput *output;
+  GfMonitorTransform new_transform;
+
+  output = gf_monitor_get_main_output (monitor);
+  new_transform = (transform + GF_MONITOR_TRANSFORM_FLIPPED -
+                   output->panel_orientation_transform) %
+                  GF_MONITOR_TRANSFORM_FLIPPED;
+
+  if (gf_monitor_transform_is_flipped (transform))
+    new_transform += GF_MONITOR_TRANSFORM_FLIPPED;
+
+  return new_transform;
+}
+
 gboolean
 gf_monitor_get_suggested_position (GfMonitor *monitor,
                                    gint      *x,
@@ -688,10 +917,14 @@ gf_monitor_get_suggested_position (GfMonitor *monitor,
 GfLogicalMonitor *
 gf_monitor_get_logical_monitor (GfMonitor *monitor)
 {
-  GfOutput *output = gf_monitor_get_main_output (monitor);
+  GfOutput *output;
+  GfCrtc *crtc;
 
-  if (output->crtc)
-    return output->crtc->logical_monitor;
+  output = gf_monitor_get_main_output (monitor);
+  crtc = gf_output_get_assigned_crtc (output);
+
+  if (crtc)
+    return crtc->logical_monitor;
   else
     return NULL;
 }
@@ -768,6 +1001,8 @@ gf_monitor_derive_current_mode (GfMonitor *monitor)
     }
 
   priv->current_mode = current_mode;
+
+  g_warn_if_fail (is_current_mode_known (monitor));
 }
 
 void
@@ -809,12 +1044,14 @@ gf_monitor_calculate_mode_scale (GfMonitor     *monitor,
                                  GfMonitorMode *monitor_mode)
 {
   GfMonitorPrivate *priv;
+  GfMonitorManager *monitor_manager;
   GfBackend *backend;
   GfSettings *settings;
   gint global_scaling_factor;
 
   priv = gf_monitor_get_instance_private (monitor);
-  backend = gf_monitor_manager_get_backend (priv->monitor_manager);
+  monitor_manager = gf_gpu_get_monitor_manager (priv->gpu);
+  backend = gf_monitor_manager_get_backend (monitor_manager);
   settings = gf_backend_get_settings (backend);
 
   if (gf_settings_get_global_scaling_factor (settings, &global_scaling_factor))
@@ -979,3 +1216,19 @@ gf_verify_monitor_mode_spec (GfMonitorModeSpec  *mode_spec,
       return FALSE;
     }
 }
+
+gboolean
+gf_monitor_has_aspect_as_size (GfMonitor *monitor)
+{
+  int width_mm;
+  int height_mm;
+
+  gf_monitor_get_physical_dimensions (monitor, &width_mm, &height_mm);
+
+  return (width_mm == 1600 && height_mm == 900) ||
+         (width_mm == 1600 && height_mm == 1000) ||
+         (width_mm == 160 && height_mm == 90) ||
+         (width_mm == 160 && height_mm == 100) ||
+         (width_mm == 16 && height_mm == 9) ||
+         (width_mm == 16 && height_mm == 10);
+}
