@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2015 Alberts Muktupāvels
+ * Copyright (C) 2015-2019 Alberts Muktupāvels
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -110,7 +110,7 @@ real_ungrab (FlashbackShell *shell,
 }
 
 static gboolean
-ungrab_accelerator (gpointer key,
+remove_accelerator (gpointer key,
                     gpointer value,
                     gpointer user_data)
 {
@@ -148,7 +148,7 @@ name_vanished_handler (GDBusConnection *connection,
   data->shell = shell;
 
   g_hash_table_foreach_remove (shell->grabbed_accelerators,
-                               ungrab_accelerator, data);
+                               remove_accelerator, data);
   g_free (data);
 
   g_bus_unwatch_name (id);
@@ -183,6 +183,28 @@ grab_accelerator (FlashbackShell *shell,
   return action;
 }
 
+static gboolean
+ungrab_accelerator (FlashbackShell *shell,
+                    guint           action,
+                    const gchar    *sender)
+{
+  const gchar *grabbed_by;
+  gboolean success;
+
+  grabbed_by = g_hash_table_lookup (shell->grabbed_accelerators,
+                                    GUINT_TO_POINTER (action));
+
+  if (g_strcmp0 (grabbed_by, sender) != 0)
+    return FALSE;
+
+  success = real_ungrab (shell, action);
+
+  if (success)
+    g_hash_table_remove (shell->grabbed_accelerators, GUINT_TO_POINTER (action));
+
+  return success;
+}
+
 static gboolean
 handle_eval (FlashbackDBusShell    *dbus_shell,
              GDBusMethodInvocation *invocation,
@@ -223,10 +245,10 @@ handle_show_osd (FlashbackDBusShell    *dbus_shell,
 }
 
 static gboolean
-handle_show_monitor_labels2 (FlashbackDBusShell    *dbus_shell,
-                             GDBusMethodInvocation *invocation,
-                             GVariant              *params,
-                             gpointer               user_data)
+handle_show_monitor_labels (FlashbackDBusShell    *dbus_shell,
+                            GDBusMethodInvocation *invocation,
+                            GVariant              *params,
+                            gpointer               user_data)
 {
   FlashbackShell *shell;
   const gchar *sender;
@@ -236,10 +258,10 @@ handle_show_monitor_labels2 (FlashbackDBusShell    *dbus_shell,
 
   g_assert (shell->monitor_manager != NULL);
 
-  flashback_monitor_labeler_show2 (shell->labeler, shell->monitor_manager,
-                                   sender, params);
+  flashback_monitor_labeler_show (shell->labeler, shell->monitor_manager,
+                                  sender, params);
 
-  flashback_dbus_shell_complete_show_monitor_labels2 (dbus_shell, invocation);
+  flashback_dbus_shell_complete_show_monitor_labels (dbus_shell, invocation);
 
   return TRUE;
 }
@@ -352,27 +374,54 @@ handle_ungrab_accelerator (FlashbackDBusShell    *dbus_shell,
                            gpointer               user_data)
 {
   FlashbackShell *shell;
-  gchar *sender;
-	gboolean success;
+  const gchar *sender;
+  gboolean success;
 
   shell = FLASHBACK_SHELL (user_data);
-  success = FALSE;
-  sender = (gchar *) g_hash_table_lookup (shell->grabbed_accelerators,
-                                          GUINT_TO_POINTER (action));
 
-  if (g_strcmp0 (sender, g_dbus_method_invocation_get_sender (invocation)) == 0)
-    {
-      success = real_ungrab (shell, action);
-
-      if (success)
-        g_hash_table_remove (shell->grabbed_accelerators, GUINT_TO_POINTER (action));
-    }
+  sender = g_dbus_method_invocation_get_sender (invocation);
+  success = ungrab_accelerator (shell, action, sender);
 
   flashback_dbus_shell_complete_ungrab_accelerator (dbus_shell, invocation, success);
 
   return TRUE;
 }
 
+static gboolean
+handle_ungrab_accelerators (FlashbackDBusShell    *dbus_shell,
+                            GDBusMethodInvocation *invocation,
+                            GVariant              *actions,
+                            gpointer               user_data)
+{
+  FlashbackShell *shell;
+  const char *sender;
+  gboolean success;
+  GVariantIter iter;
+  GVariant *child;
+
+  shell = FLASHBACK_SHELL (user_data);
+
+  sender = g_dbus_method_invocation_get_sender (invocation);
+  success = TRUE;
+
+  g_variant_iter_init (&iter, actions);
+  while ((child = g_variant_iter_next_value (&iter)))
+    {
+      guint action;
+
+      g_variant_get (child, "u", &action);
+      g_variant_unref (child);
+
+      success &= ungrab_accelerator (shell, action, sender);
+    }
+
+  flashback_dbus_shell_complete_ungrab_accelerators (dbus_shell,
+                                                     invocation,
+                                                     success);
+
+  return TRUE;
+}
+
 static void
 name_appeared_handler (GDBusConnection *connection,
                        const gchar     *name,
@@ -392,8 +441,8 @@ name_appeared_handler (GDBusConnection *connection,
                     G_CALLBACK (handle_focus_search), shell);
   g_signal_connect (skeleton, "handle-show-osd",
                     G_CALLBACK (handle_show_osd), shell);
-  g_signal_connect (skeleton, "handle-show-monitor-labels2",
-                    G_CALLBACK (handle_show_monitor_labels2), shell);
+  g_signal_connect (skeleton, "handle-show-monitor-labels",
+                    G_CALLBACK (handle_show_monitor_labels), shell);
   g_signal_connect (skeleton, "handle-hide-monitor-labels",
                     G_CALLBACK (handle_hide_monitor_labels), shell);
   g_signal_connect (skeleton, "handle-focus-app",
@@ -406,6 +455,8 @@ name_appeared_handler (GDBusConnection *connection,
                     G_CALLBACK (handle_grab_accelerators), shell);
   g_signal_connect (skeleton, "handle-ungrab-accelerator",
                     G_CALLBACK (handle_ungrab_accelerator), shell);
+  g_signal_connect (skeleton, "handle-ungrab-accelerators",
+                    G_CALLBACK (handle_ungrab_accelerators), shell);
 
   flashback_dbus_shell_set_mode (skeleton, "");
   flashback_dbus_shell_set_overview_active (skeleton, FALSE);
@@ -437,6 +488,14 @@ flashback_shell_finalize (GObject *object)
       shell->bus_name = 0;
     }
 
+  if (shell->iface != NULL)
+    {
+      g_dbus_interface_skeleton_unexport (shell->iface);
+
+      g_object_unref (shell->iface);
+      shell->iface = NULL;
+    }
+
   if (shell->grabbed_accelerators)
     {
       g_hash_table_destroy (shell->grabbed_accelerators);
